/*
Created: 22/07/2011
*/

#pragma once
#ifndef BTTRIANGLEMESHEX_H__
#define BTTRIANGLEMESHEX_H__

// Made just for this demo to allow loading more than one (sub)mesh
// The implementation uses STL, not this header.

#ifdef _WIN32 //needed for glut.h
#include <windows.h>
#endif
#include "GlutStuff.h"	// Just for GLuint...

//#define INCLUDE_MESHES_FROM_APPGIMPACTTESTDEMO	// These meshes are probably GPL licensed, so not embeddable in commercial applications...
#define LOAD_MESHES_FROM_STRIDING_MESH_INTERFACE    // Needs include folder: "Bullet/Extra/Serialize" and needs link to: BulletWorldImporter.lib

// This class does very little: just stores a list of all the DLs, so we can delete them all later...
struct DisplayListHelper	{
	public:
	static void ClearArray();
	static void Add(GLuint dl);
	static bool IsValid(GLuint dl);
	
	protected:
	DisplayListHelper()	{}
	static btAlignedObjectArray< GLuint > g_displayListArray;
	
};

struct TextureHelper	{
	public:
	static void ClearArray();
	static void Add(GLuint id);
	static bool IsValid(GLuint id);
	static GLuint Get(const char* textureFileName);	// 0 = no texture

	protected:
	TextureHelper()	{}
	static btAlignedObjectArray< GLuint > g_textureArray;
	//friend class GlutDemoApplicationEx;	//remove me
};

struct Material	{
	float amb[4];
	float dif[4];
	float spe[4];
	float shi;	// In [0,1], not in [0,128]
	float emi[4];
	Material();	
	Material(float r,float g,float b,bool useSpecularColor = true,float ambFraction=0.4f,float speAddon=0.4f,float shininessIn01=(4.0f/128.0f));	
	// TODO: add ctrs for known colors
	void bind() const;	// openGL method: btTriangleMesh.h/cpp shouldn't include gl stuff, we'll define it in another .cpp file.		
	void lighten(float q=0.2f);
	void darken(float q=0.2f);
	void clamp();	// all in [0,1]
	void reset();
	//Uses printf
	void displayDebugInfo() const;
};

struct btVector2	{
	protected:
	btScalar m_x;
	btScalar m_y;
	public:
	btVector2() {}
	btVector2(btScalar _x,btScalar _y) : m_x(_x),m_y(_y) {}
	
	const btScalar& x() const {return m_x;}
	const btScalar& y() const {return m_y;}
	
	void setX(btScalar x) {m_x=x;}
	void setY(btScalar y) {m_y=y;}
	void setValue(btScalar x,btScalar y) {m_x=x;m_y=y;}
};

class btTriangleMeshEx	{
	protected:
	typedef class btTriangleIndexVertexArray* btTriangleIndexVertexArrayPtr;
	
	public:
	btTriangleMeshEx();
	
	virtual ~btTriangleMeshEx();

	// Auto cast to btTriangleIndexVertexArray*... a bit dangerous, but I'm very lazy...
	operator btTriangleIndexVertexArrayPtr();
	
	// Mandatory to clean the mesh. Otherwise we can load multiple meshes and they will be added to the
	// meshes already present
	void reset();
	
	// verts, inds, texcoords and materials are imported: supports one vertex array and many index arrays (subparts)
	bool loadFromOBJFile(const char* filePath,bool invertYZ=false);

	// Very simplificated method made just for Bullet: only verts and inds are imported: supports one vertex array and one index array (only one subpart)
	bool loadFromOFFFile(const char* filePath,bool invertYZ=false);

	#ifdef LOAD_MESHES_FROM_STRIDING_MESH_INTERFACE
	bool loadFromMeshInterface(const class btStridingMeshInterface* smi,bool invertYZ=false,const class btVector3* pOptionalScaling=NULL,const class btTransform* pOptionalTransform=NULL);
	// The following method should work as long as a btStridingMeshInterface is present in the collision shape (i.e. only some concave shapes work)
	bool loadFromMeshInterfaceInsideBulletCollisionShape(const btCollisionShape* shape,bool invertYZ=false);
	// bcsFilePath can be the path to a .bullet file with just one btCollisionShape in it
	// The following method should work as long as a btStridingMeshInterface is present in the collision shape (i.e. only some concave shapes work)
	bool loadFromMeshInterfaceInsideBCSFile(const char* bcsFilePath,bool invertYZ=false);	
	#endif //LOAD_MESHES_FROM_STRIDING_MESH_INTERFACE
	
	// Debug only. Only verts and inds are imported: supports one vertex array and one index array. Made just to save obj meshes with HACD problems.
	bool saveToOFFFile(const char* filePath);

	// OBJ or OFF only...
	bool loadFromFile(const char* filePath,bool invertYZ=false);

	#ifdef INCLUDE_MESHES_FROM_APPGIMPACTTESTDEMO
	// Default meshes (embedded in source code)
	bool loadDefaultBunnyMesh();	
	bool loadDefaultTorusMesh();
	#endif //INCLUDE_MESHES_FROM_APPGIMPACTTESTDEMO
		
	// We could add more methods like this to build other default meshes... here axis is: +-1 = +-X,		+-2 = +-Y		or 		+-3 = +-Z.
	bool loadDefaultPlaneMesh(int axis=2,btScalar halfLength=btScalar(1.0),btScalar halfHeight=btScalar(1.0),btScalar texRepX=btScalar(1.0),btScalar texRepY=btScalar(1.0),btScalar texOffX=btScalar(0.0),btScalar texOffY=btScalar(0.0),bool twoSided = false);
	bool loadDefaultBoxMesh(const btVector3& halfExtents,bool useTextures=false,bool useOneMaterialForEachFace=false,btScalar texRepX=btScalar(1.0),btScalar texRepY=btScalar(1.0),btScalar texOffX=btScalar(0.0),btScalar texOffY=btScalar(0.0),unsigned short numOfFacesToCreate=6);
	
	// optionalMeshIndex!=-1 has some meaning only if more than one mesh is loaded at the same time:
        int getNumSubParts(int optionalMeshIndex=-1) const;
        int getBaseSubPartIndex(int optionalMeshIndex=-1) const;	// returns the global subPartIndex of the mesh number meshIndex (the subPart containing the vertex array referenced by all the other subparts of this mesh)
	btAlignedObjectArray< int > getSubpartIndices(int optionalMeshIndex=-1) const;	// returns an array of global subPartIndices of the mesh number meshIndex
	
	// The following actions are not persistent, but instantaneous.
	// optionalSubpartIndex!=-1 has undefined results if meshes[optionalSubpartIndex] does not own his own vertex array
	btVector3 center(int  optionalMeshIndex=-1);	// (returns the old center)
	btVector3 setHeight(btScalar height,int  optionalMeshIndex=-1);		// Y axis (returns the applied scaling factor)
	btVector3 setWidth(btScalar width,int  optionalMeshIndex=-1);		// X axis (returns the applied scaling factor)
	btVector3 setLength(btScalar length,int  optionalMeshIndex=-1);		// Z axis (returns the applied scaling factor)	
	btVector3 setFullExtents(const btVector3& extents,int  optionalMeshIndex=-1); // (returns the applied scaling factor)
	btVector3 setHalfExtents(const btVector3& extents,int  optionalMeshIndex=-1); // (returns the applied scaling factor)
	void shift(const btVector3& shiftFactor,int optionalMeshIndex=-1);
	void scale(const btVector3& scaling,int optionalMeshIndex=-1);
	void rotate(const btMatrix3x3& m,int optionalMeshIndex=-1);
	void rotate(const btQuaternion& q,int optionalMeshIndex=-1);
	void transform(const btTransform& T,int optionalMeshIndex=-1); 
	
	const btVector3 getAabbHalfExtents(int optionalMeshIndex=-1) const;
	const btVector3 getAabbFullExtents(int optionalMeshIndex=-1) const;
	const btVector3 getCenterPoint(int optionalMeshIndex=-1) const;
	

	//very big memory copy... slow
	btTriangleMeshEx(const btTriangleMeshEx& c);
	//very big memory copy... slow
	const btTriangleMeshEx& operator=(const btTriangleMeshEx& c);

	//Uses printf
	void displayDebugInfo() const;
	
	
		
	
	struct btMeshSubPart {
		// Very space intensive...
		btAlignedObjectArray < btVector3 > verts;	// if size()==0, inds refers to useVertexArrayInSubpartIndex if != -1 ot to the last "valid" verts in the previous subparts.
		btAlignedObjectArray < int > inds;
		int useVertexArrayInSubpartIndex;			// Used only if verts.size()==0, useVertexArrayInSubpartIndex must be lesser than the current subpart index

		btAlignedObjectArray < btVector2 > texcoords;	// if size()==0, texinds refers to useTexcoordArrayInSubpartIndex if != -1 ot to the last "valid" texcoords in the previous subparts.
		btAlignedObjectArray < int > tinds;
		int useTexcoordArrayInSubpartIndex;			// Used only if verts.size()==0, useVertexArrayInSubpartIndex must be lesser than the current subpart index
		bool useIndsAsTexcoordInds;					// If false (default) tinds.size() == inds.size(). If true verts.size() == texcoords.size() (each belonging to this or the used subPart)
		
		Material material;
		GLuint textureId;
		char textureFilePath[1025];
		
		btMeshSubPart();
		btMeshSubPart(int _useVertexArrayInSubpartIndex,int _useTexcoordArrayInSubpartIndex=-1,bool _useIndsAsTexcoordInds=false);
		protected:
		void reset();
	};
	const btAlignedObjectArray<btMeshSubPart>& getMeshSubParts() const {return meshSubParts;}	
	const btMeshSubPart* getMeshSubPart(int i) const {return ((i>=0 && i<meshSubParts.size()) ? &meshSubParts[i] : NULL);}
	const Material* getMeshSubPartMaterial(int i) const {return ((i>=0 && i<meshSubParts.size()) ? &meshSubParts[i].material : NULL);}
	Material* getMeshSubPartMaterial(int i) {return ((i>=0 && i<meshSubParts.size()) ? &meshSubParts[i].material : NULL);}
	void setMeshSubPartMaterial(int i,const Material& mat) {if (i>=0 && i<meshSubParts.size())  meshSubParts[i].material = mat;}
        GLuint getMeshSubPartTextureId(int i) const {return ((i>=0 && i<meshSubParts.size()) ? meshSubParts[i].textureId : 0);}
	void setMeshSubPartTextureId(int i,GLuint textureId) {if (i>=0 && i<meshSubParts.size())  meshSubParts[i].textureId = textureId;}
		
        int getNumMeshSubParts() const {return meshSubParts.size();}	// Equivalent to getNumSubParts(-1);

	public:
	// This should be protected...
	void createMeshFootPrint(class btTriangleMeshExFootPrint*& p,const btVector3& scaling=btVector3(1,1,1),const bool useFlatNormals=false,const btAlignedObjectArray< const Material* >* pOptionalMaterialOverrideArray=NULL,const btAlignedObjectArray< GLuint >* pOptionalTextureOverrideArray=NULL,const btAlignedObjectArray< int >* pOptionalSubmeshIndexArray=NULL) const;	//Allocates and fills it
	//static void DeleteMeshFootPrint(btTriangleMeshExFootPrint*& p);	//Deletes it

	protected:
	btAlignedObjectArray<btMeshSubPart> meshSubParts;	// I should use <btMeshSubPart*>, but I want to keep things simple
	btTriangleIndexVertexArray* tiva;
	bool isTivaDirty;

	btVector3 GetAabbHalfExtents(btVector3* pOptionalCenterPointOut=NULL,int optionalMeshIndex=-1) const;
	
	bool glDrawMeshSubPart(int i,const btVector3& scaling = btVector3(1,1,1), const bool useFlatNormals=false) const;
	public:
	struct btMeshSubPartFootPrint	{
		GLuint dl;	// display list
		Material material;
		GLuint textureId;
		btMeshSubPartFootPrint() : dl(0),textureId(0) {}
	};
	protected:
	friend class btTriangleMeshExFootPrint;	
	//friend class btTriangleMeshExFootPrint;
	//static btAlignedObjectArray<btTriangleMeshExFootPrint*> g_btTriangleMeshExFootPrints;
};

class btTriangleMeshExFootPrint : public btAlignedObjectArray<btTriangleMeshEx::btMeshSubPartFootPrint>	{
	protected:
	btTriangleMeshExFootPrint() {}
	virtual ~btTriangleMeshExFootPrint() {}
	friend class btTriangleMeshEx;
	public:
	struct Overrides {
		btVector3 offset;
		btVector3 scaling;
		btAlignedObjectArray< const Material*> materialArray;
		btAlignedObjectArray< GLuint > textureArray;
		bool disableSpecularOnTexture;
                bool avoidGlNormalizeCallsWhenScaling;      // Better leave it to false
                Overrides() : offset(0,0,0),scaling(1,1,1),disableSpecularOnTexture(false),avoidGlNormalizeCallsWhenScaling(false) {}
		void displayDebugInfo() const;
		void reset(); 
		Overrides(const Overrides& ovr);
		const Overrides& operator=(const Overrides& ovr);	// Unluckily this needs a copy of all struct params in the .cpp file ( because it assumes BT_ALLOW_ARRAY_COPY_OPERATOR just optional )
		
	};
	void glDraw(const btTriangleMeshExFootPrint::Overrides* ovr=NULL,const btTriangleMeshExFootPrint::Overrides* ovrFallback=NULL,const bool useAmbientLightOnly = false) const;
	
	public:
	struct Helper	{
		public:
		static void ClearArray();
		protected:
		static void Add(const btTriangleMeshExFootPrint* dl);
		static bool IsValid(const btTriangleMeshExFootPrint* dl);	
			
		protected:
		Helper();
		static btAlignedObjectArray< const btTriangleMeshExFootPrint* > g_triangleMeshExFootPrintArray;
		friend class btTriangleMeshEx;
	};
};


#endif //BTTRIANGLEMESHEX_H__



